'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.traverse = exports.toStackOf = exports.toListOf = exports.index = exports.field = exports.contains = undefined;

var _lens = require('./lens');

var _Applicative = require('./lib/Applicative');

var _immutable = require('immutable');

exports.contains = contains;
exports.field = field;
exports.index = index;
exports.toListOf = toListOf;
exports.toStackOf = toStackOf;
exports.traverse = traverse;

/* lenses */

function field(name) {
  return (0, _lens.lens)(function (obj) {
    return obj.get(name);
  }, function (obj, val) {
    return obj.set(name, val);
  });
}

function contains(val) {
  return (0, _lens.lens)(function (obj) {
    return obj.has(val);
  }, function (obj, b) {
    return b ? obj.add(val) : obj.remove(val);
  });
}

/* traversals */

function index(idx) {
  return function (f) {
    return function (pure, obj) {
      if (obj.has(idx)) {
        return f(pure, obj.get(idx)).map(function (updatedValue) {
          // Optimize update for certain types
          if (typeof obj.set === 'function') {
            return obj.set(idx, updatedValue);
          } else {
            // This works for any Iterable, but might not give the most efficient update.
            return obj.map(function (v, k) {
              return k === idx ? updatedValue : v;
            });
          }
        });
      } else {
        return pure(obj);
      }
    };
  };
}

/*
 * You can use the `traverse` function to get a `Traversal` from any
 * `Traversable` value.
 *
 * A `Traversal` is a specialized lens: it can act as a setter,
 * but cannot be used as a getter with the `get` function.
 *
 * For example:
 *
 *     var xs = List([1,2,3])
 *
 *     over(traverse, x => x * 2, xs)
 *
 *     assert(is( xs, List([2,4,6]))
 *
 * You can use `lookup` with a `Traversal` to get values out of a structure
 * wrapped in a `Maybe` result.
 * For example, using `index`, which is a traversal:
 *
 *     import { Just, Nothing } from 'lens/src/Maybe'
 *
 *     var x = lookup(index(1), xs)
 *
 *     assert( x instanceof Just )
 *     assert( x.value === 2 )
 *
 *     var y = lookup(index(9), xs)
 *
 *     assert( y instanceof Nothing )
 *     assert( y.value === undefined )
 *
 */

function traverse(f) {
  return function (pure, obj) {
    if (obj instanceof _immutable.Iterable.Keyed || obj instanceof _immutable.Collection.Keyed) {
      return traverseKeyedIterable(f)(pure, obj);
    } else if (obj instanceof _immutable.Iterable) {
      return traverseIterable(f)(pure, obj);
    } else {
      throw new TypeError("No `traverse` implementation for " + nameOfType(obj));
    }
  };
}

function traverseIterable(f) {
  return function (pure, obj) {
    var cons = consImpl(obj);
    var emptyColl = obj.take(0);
    return obj.reduceRight(function (ys, x) {
      return f(pure, x).map(cons).ap(ys);
    }, pure(emptyColl));
  };
}

function traverseKeyedIterable(f) {
  return function (pure, obj) {
    var cons = consKeyedImpl(obj);
    var emptyColl = obj.take(0);
    return obj.reduceRight(function (ys, x, key) {
      return f(pure, x).map(cons).ap(pure(key)).ap(ys);
    }, pure(emptyColl));
  };
}

function consImpl(coll) {
  if (coll instanceof _immutable.List || coll instanceof _immutable.Stack) {
    return function (val) {
      return function (coll) {
        return coll.unshift(val);
      };
    };
  } else if (coll instanceof _immutable.Set) {
    return function (val) {
      return function (coll) {
        return coll.add(val);
      };
    };
  } else {
    var emptyColl = coll.take(0);
    return function (val) {
      return function (coll) {
        return emptyColl.concat(val).concat(coll);
      };
    };
  }
}

function consKeyedImpl(coll) {
  if (coll instanceof Map) {
    return function (val) {
      return function (key) {
        return function (coll) {
          return coll.set(key, val);
        };
      };
    };
  } else {
    var emptyColl = coll.take(0);
    return function (val) {
      return function (key) {
        return function (coll) {
          return emptyColl.concat([[key, val]]).concat(coll);
        };
      };
    };
  }
}

function nameOfType(obj) {
  return obj && obj.constructor && obj.constructor.name ? obj.constructor.name : String(obj);
}

/* traversing */

function toListOf(l, obj) {
  return (0, _lens.foldrOf)(l, function (x, xs) {
    return xs.unshift(x);
  }, (0, _immutable.List)(), obj);
}

/*
 * Constructing a stack using foldr might be more efficient than constructing
 * a list, since Immutable's Stack is implemented as a linked list, while List
 * is implemented as a tree-backed vector.
 */
function toStackOf(l, obj) {
  return (0, _lens.foldrOf)(l, function (x, xs) {
    return xs.unshift(x);
  }, (0, _immutable.Stack)(), obj);
}
